#include <sqlite3.h>
#include <stdio.h>
#include <libconfig.h>

/*
 * Copyright 2010 Sven Vermeulen.
 * Subject to the GNU Public License, version 3.
 */
 
#ifndef _CVETYPES
#define _CVETYPES

struct arguments {
	char * args;
	char * binlist;
	char * cvedata;
	char * singlefile;
	char * datafile;
	int parsebin;
	int loadcve;
	int runcheck;
	int hassinglefile;
	int hasdatafile;
	int initdatabases;
	int docsvoutput;
	int doshowinstalled;
	int doshowinstalledfiles;
};

struct cpe_data {
	char part;
	char vendor[64];
	char product[64];
	char version[64];
	char update[64];
	char edition[64];
	char language[64];
};

struct workstate {
#ifdef _USE_SQLITE3
	sqlite3 * matchdb;
	sqlite3 ** localdb;
#endif
	struct arguments * arg;
	FILE * binlist;
	FILE * datafile;
	char * currentdir;
	char * currentfile;
	config_t * cfg;
	struct cpe_data cpebuffer;
	void ** resultlist;
	int numresults;
	int rc;
};

struct versiongather_data {
	char filepart[64];
	int gathertype;
	char filematch[128];
	char versionexpression[256];
};

/***********************************************************************************************
 * CPE related definitions
 ***********************************************************************************************/

// cpe_to_string - Convert the selected cpe_data structure to a string
void cpe_to_string(char * buffer, int buffsize, struct cpe_data cpe);

// string_to_cpe - Convert the selected cpe string (buffer) to a structure
void string_to_cpe(struct cpe_data * cpe, char * buffer);

// cve_to_vars - Convert the cve identifier (string) to a year/sequence combination
int cve_to_vars(int * year, int * sequence, char * cveId);

// show_potential_vulnerabilities - Show the potential vulnerability matches
void show_potential_vulnerabilities(struct workstate * ws, int cveyear, int cvenum, const char * filename, struct cpe_data cpe, int versiononly);

// show_installed_software - Show the installed software
void show_installed_software(struct workstate * ws, const char * vendor, const char * product, const char * version, const char * update, const char * edition, const char * language, int numfiles, const char ** files);

// clear_resultlist - Clear workstate result list
void clear_resultlist(struct workstate * ws);

#endif
