#include "impl.h"

/*
 * Copyright 2010 Sven Vermeulen.
 * Subject to the GNU Public License, version 3.
 */
 
/***********************************************************************************************
 * Helper functions for the sqlite3 database manipulations
 ***********************************************************************************************/

sqlite3 * get_local_db(struct workstate * ws, char part, int len) {
	if (part == 'a') {
		return ws->localdb[len];
	} else if (part == 'h') {
		return ws->localdb[64+len];
	} else if (part == 'o') {
		return ws->localdb[2*64+len];
	};

	return NULL;
};

int get_db_intresult(void * cbobj, int argc, char **argv, char **azColName) {
	int i = 0;
	struct workstate * ws = (struct workstate *) cbobj;

	for(i=0;i<argc;i++) {
		// Found result. Storing in workstate variable
		int ret = atoi(argv[i]);
		ws->rc=ret;
		return 0;
	};

	return 0;
};

int get_cpe_dataresult(void * cbobj, int argc, char **argv, char **azColName) {
	int i = 0;
	struct workstate * ws = (struct workstate *) cbobj;

	for(i = 0; i < argc; i++) {
		if (strcmp(azColName[i], "cpepart") == 0) {
			ws->cpebuffer.part = argv[i][0];
			continue;
		}
		if (strcmp(azColName[i], "cpevendor") == 0) {
			strncpy(ws->cpebuffer.vendor, argv[i], 64);
			continue;
		}
		if (strcmp(azColName[i], "cpeproduct") == 0) {
			strncpy(ws->cpebuffer.product, argv[i], 64);
			continue;
		}
		if (strcmp(azColName[i], "cpeversion") == 0) {
			strncpy(ws->cpebuffer.version, argv[i], 64);
			continue;
		}
		if (strcmp(azColName[i], "cpeupdate") == 0) {
			strncpy(ws->cpebuffer.update, argv[i], 64);
			continue;
		}
		if (strcmp(azColName[i], "cpeedition") == 0) {
			strncpy(ws->cpebuffer.edition, argv[i], 64);
			continue;
		}
		if (strcmp(azColName[i], "cpelanguage") == 0) {
			strncpy(ws->cpebuffer.language, argv[i], 64);
			continue;
		};
	};

	ws->rc = 0;
	return 0;
};

int get_filelist(void * cbobj, int argc, char **argv, char **azColName) {
	int i = 0;
	struct workstate * ws = (struct workstate *) cbobj;
	int current = ws->rc;
	char basedir[64];
	char filename[64];
	char * fullfilename;

	for (i = 0; i < argc; i++) {
		if (strcmp(azColName[i], "basedir") == 0) {
			strncpy(basedir, argv[i], 64);
			continue;
		}
		if (strcmp(azColName[i], "filename") == 0) {
			strncpy(filename, argv[i], 64);
			continue;
		};
	};

	fullfilename = (char *) calloc(128, sizeof(char));
	sprintf(fullfilename, "%s/%s", basedir, filename);
	ws->resultlist[current] = fullfilename;
	ws->rc = current+1;

	return 0;
};

int get_cpe_data(sqlite3 * db, char * stmt, struct workstate * ws) {
	int rc = 0;
	char * errmsg;

	ws->rc = 0;
	rc = sqlite3_exec(db, stmt, get_cpe_dataresult, ws, &errmsg);
	if (rc) {
		fprintf(stderr, "Failed to execute CPE-returning statement against database: %s\n", errmsg);
		sqlite3_free(errmsg);
	};

	return ws->rc;
};

int get_int_value(sqlite3 * db, char * stmt, struct workstate * ws) {
	int rc = 0;
	char * errmsg;

	ws->rc = 0;
	rc = sqlite3_exec(db, stmt, get_db_intresult, ws, &errmsg);
	if (rc) {
		fprintf(stderr, "Failed to execute INT-returning statement against database: %s\n", errmsg);
		sqlite3_free(errmsg);
	};

	return ws->rc;
};

int get_cpelist(void * cbobj, int argc, char **argv, char **azColName) {
	int i = 0;
	struct workstate * ws = (struct workstate *) cbobj;
	char part = '\0';
	char stmt[1024];
	int length = 0;
	int cpeid = 0;
	int showfiles = ws->rc;
	struct cpe_data cpedata;
	char * errmsg;

	if (argc != 3)
		return 1;

	for (i = 0; i < argc; i++) {
		if (strcmp(azColName[i], "cpepart") == 0) {
			part = argv[i][0];
			continue;
		}
		if (strcmp(azColName[i], "cpevendorlength") == 0) {
			length = atoi(argv[i]);
			continue;
		}
		// This one must be at the end, 'cause it always matches the previous hits as well
		if (strcmp(azColName[i], "cpe") == 0) {
			cpeid = atoi(argv[i]);
			continue;
		}
	};
	
	zero_string(stmt, 1024);
	sprintf(stmt, "select cpepart, cpevendor, cpeproduct, cpeversion, cpeupdate, cpeedition, cpelanguage from tb_cpe_%c_%d where cpeid = %d;", part, length, cpeid);
	i = get_cpe_data(get_local_db(ws, part, length), stmt, ws);
	if (i != 0)
		return i;
	cpedata = ws->cpebuffer;
	ws->numresults = 0;
	if (showfiles) {
		// Step 1 - Get number of hits
		zero_string(stmt, 1024);
		sprintf(stmt, "SELECT count(rowid) FROM tb_binmatch WHERE cpepart = \'%c\' AND cpevendorlength = %d AND cpe = %d;", part, length, cpeid);
		i = get_int_value(ws->localdb[0], stmt, ws);
		ws->resultlist = (void **) calloc(i, sizeof(char *));
		ws->numresults = i;
		ws->rc = 0;
		// Step 2 - Get list of files
		zero_string(stmt, 1024);
		sprintf(stmt, "SELECT basedir, filename FROM tb_binmatch WHERE cpepart = \'%c\' AND cpevendorlength = %d AND cpe = %d;", part, length, cpeid);
		i = sqlite3_exec(ws->localdb[0], stmt, get_filelist, ws, &errmsg);
		if (i) {
			fprintf(stderr, "Failed to get list of files for a particular software/version set: %s\n", errmsg);
			sqlite3_free(errmsg);
			return i;
		};
	};
	show_installed_software(ws, cpedata.vendor, cpedata.product, cpedata.version, cpedata.update, cpedata.edition, cpedata.language, ws->numresults, (const char **) ws->resultlist);

	if (showfiles) {
		clear_resultlist(ws);
		free(ws->resultlist);
	};

	return i;
};


void run_statement(struct workstate * ws, sqlite3 * db, char * stmt) {
	int rc = 0;
	char * errmsg;

	rc = sqlite3_exec(db, stmt, NULL, NULL, &errmsg);
	if (rc) {
		fprintf(stderr, "Failed to run statement: %s\n", errmsg);
		sqlite3_free(errmsg);
	};
};

void run_statement_alldb(struct workstate * ws, char * stmt) {
	int i = 0;

	for (i = 0; i < 3*64+1; i++) {
		run_statement(ws, ws->localdb[i], stmt);
	};
};

/**
 * Clear the version database.
 */
void dbimpl_clear_versiondatabase(struct workstate * ws) {
	char stmt[1024];

	sprintf(stmt, "delete from tb_binmatch;");

	run_statement(ws, ws->localdb[0], stmt);
};

/**
 * Clear the version gathering data
 */
void dbimpl_clear_versiondata(struct workstate * ws) {
	char stmt[1024];

	sprintf(stmt, "delete from tb_versionmatch;");

	run_statement(ws, ws->matchdb, stmt);

	sprintf(stmt, "delete from tb_cpe;");

	run_statement(ws, ws->matchdb, stmt);
};

/**
 * Run updates on the database (due to cvechecker upgrades or 
 * fixes)
 */
void run_upgrade_fixes(struct workstate * ws) {
	int rc = 0;
	int i;
	char stmt[1024];


	/**
	 * 1 - Add tables tb_cpe_a_64, tb_cpe_h_64 and tb_cpe_o_64 if they don't exist yet
	 *
	 * If count(*) returns 0, then there are no tables in the database, so we can create.
	 */
	sprintf(stmt, "select count(*) from sqlite_master;");
	rc = get_int_value(get_local_db(ws, 'a', 64), stmt, ws);
	if (rc == 0) {
		fprintf(stderr, "I am missing the tables in a64 (tb_cpe_a_64). This is to be expected if you upgraded cvechecker from 0.5 or lower. I will now create tb_cpe_a_64 for you, no further actions are needed.\n");
		zero_string(stmt, 1024);
		sprintf(stmt, "CREATE TABLE tb_cpe_a_64 (cpeid integer primary key, cpepart char(1), cpevendor char(64), cpeproduct char(64), cpeversion char(64), cpeupdate char(64), cpeedition char(64), cpelanguage char(64));");
		run_statement(ws, get_local_db(ws, 'a', 64), stmt);
	};
	sprintf(stmt, "select count(*) from sqlite_master;");
	rc = get_int_value(get_local_db(ws, 'h', 64), stmt, ws);
	if (rc == 0) {
		fprintf(stderr, "I am missing the tables in h64 (tb_cpe_h_64). This is to be expected if you upgraded cvechecker from 0.5 or lower. I will now create tb_cpe_h_64 for you, no further actions are needed.\n");
		zero_string(stmt, 1024);
		sprintf(stmt, "CREATE TABLE tb_cpe_h_64 (cpeid integer primary key, cpepart char(1), cpevendor char(64), cpeproduct char(64), cpeversion char(64), cpeupdate char(64), cpeedition char(64), cpelanguage char(64));");
		run_statement(ws, get_local_db(ws, 'h', 64), stmt);
	};
	sprintf(stmt, "select count(*) from sqlite_master;");
	rc = get_int_value(get_local_db(ws, 'o', 64), stmt, ws);
	if (rc == 0) {
		fprintf(stderr, "I am missing the tables in o64 (tb_cpe_o_64). This is to be expected if you upgraded cvechecker from 0.5 or lower. I will now create tb_cpe_o_64 for you, no further actions are needed.\n");
		zero_string(stmt, 1024);
		sprintf(stmt, "CREATE TABLE tb_cpe_o_64 (cpeid integer primary key, cpepart char(1), cpevendor char(64), cpeproduct char(64), cpeversion char(64), cpeupdate char(64), cpeedition char(64), cpelanguage char(64));");
		run_statement(ws, get_local_db(ws, 'o', 64), stmt);
	};

	/**
	 * 2 - Add indexes, performance
	 */
	sprintf(stmt, "select count(rowid) from sqlite_master where name = 'binmatchidx';");
	rc = get_int_value(ws->localdb[0], stmt, ws);
	if (rc == 0) {
		fprintf(stderr, "I am missing the index binmatchidx. This is to be expected if you upgraded cvechecker from 0.5 or lower. I will now create binmatchidx for you, no further actions are needed.\n");
		zero_string(stmt, 1024);
		sprintf(stmt, "CREATE INDEX binmatchidx on tb_binmatch (cpe, cpepart, cpevendorlength);");
		run_statement(ws, ws->localdb[0], stmt);
	};
	for (i = 1; i <= 64; i++) {
		sprintf(stmt, "select count(rowid) from sqlite_master where name = 'cpe_a_%d_idx';", i);
		rc = get_int_value(get_local_db(ws, 'a', i), stmt, ws);
		if (rc == 0) {
			fprintf(stderr, "I am missing the index cpe_a_%d_idx. This is to be expected if you upgraded cvechecker from 0.5 or lower. I will now create cpe_a_%d_idx for you, no further actions are needed.\n", i, i);
			zero_string(stmt, 1024);
			sprintf(stmt, "CREATE INDEX cpe_a_%d_idx on tb_cpe_a_%d (cpevendor, cpeproduct, cpeversion, cpeid, cpeedition, cpeupdate, cpelanguage);", i, i);
			run_statement(ws, get_local_db(ws, 'a', i), stmt);
		};

		sprintf(stmt, "select count(rowid) from sqlite_master where name = 'cpe_h_%d_idx';", i);
		rc = get_int_value(get_local_db(ws, 'h', i), stmt, ws);
		if (rc == 0) {
			fprintf(stderr, "I am missing the index cpe_h_%d_idx. This is to be expected if you upgraded cvechecker from 0.5 or lower. I will now create cpe_h_%d_idx for you, no further actions are needed.\n", i, i);
			zero_string(stmt, 1024);
			sprintf(stmt, "CREATE INDEX cpe_h_%d_idx on tb_cpe_h_%d (cpevendor, cpeproduct, cpeversion, cpeid, cpeedition, cpeupdate, cpelanguage);", i, i);
			run_statement(ws, get_local_db(ws, 'h', i), stmt);
		};

		sprintf(stmt, "select count(rowid) from sqlite_master where name = 'cpe_o_%d_idx';", i);
		rc = get_int_value(get_local_db(ws, 'o', i), stmt, ws);
		if (rc == 0) {
			fprintf(stderr, "I am missing the index cpe_o_%d_idx. This is to be expected if you upgraded cvechecker from 0.5 or lower. I will now create cpe_o_%d_idx for you, no further actions are needed.\n", i, i);
			zero_string(stmt, 1024);
			sprintf(stmt, "CREATE INDEX cpe_o_%d_idx on tb_cpe_o_%d (cpevendor, cpeproduct, cpeversion, cpeid, cpeedition, cpeupdate, cpelanguage);", i, i);
			run_statement(ws, get_local_db(ws, 'o', i), stmt);
		};
	};
};

/**
 * Load the databases into the workstate
 *
 * The databases that are needed for the SQLite3 implementation are loaded in
 * the workstate variable. 
 */
int dbimpl_load_databases(struct workstate * ws) {
	const config_setting_t * localdb;
	const config_setting_t * globaldb;
	int rc = 0;
	int i = 0;
	char buffer[512];
	char buffer2[512];

	/*
	 * Load the set of databases
	 */
	localdb = config_lookup(ws->cfg, "localdb");
	if (localdb == NULL) {
		fprintf(stderr, "Configuration file does not contain localdb directive.\n");
		return 1;
	};
	rc = strlen(config_setting_get_string(localdb));
	if (rc > 500) {
		fprintf(stderr, "Configuration files 'localdb' directive cannot exceed 500 characters\n");
		return 1;
	};
	strncpy(buffer, config_setting_get_string(localdb), rc);
	buffer[rc] = 0x00;
	strcat(buffer, "/");

	sprintf(buffer2, "%smain.db", buffer);
	rc = sqlite3_open(buffer2, &(ws->localdb[0]));
	if (rc) {
		fprintf(stderr, "Can't open database %s: %s\n", buffer2, sqlite3_errmsg(ws->localdb[0]));
		sqlite3_close(ws->localdb[0]);
		return rc;
	};
	run_statement(ws, ws->localdb[0], "PRAGMA cache_size=10000;");
	run_statement(ws, ws->localdb[0], "PRAGMA synchronous=OFF;");

	// a-part
	for (i = 1; i <= 64; i++) {
		sprintf(buffer2, "%sa%d.db", buffer, i);
		rc = sqlite3_open(buffer2, &(ws->localdb[i]));
		if (rc) {
			fprintf(stderr, "Can't open database %s: %s\n", buffer2, sqlite3_errmsg(ws->localdb[i]));
			sqlite3_close(ws->localdb[i]);
			return rc;
		};
		run_statement(ws, ws->localdb[i], "PRAGMA cache_size=10000;");
		run_statement(ws, ws->localdb[i], "PRAGMA synchronous=OFF;");
	};
	// h-part
	for (i = 1; i <= 64; i++) {
		sprintf(buffer2, "%sh%d.db", buffer, i);
		rc = sqlite3_open(buffer2, &(ws->localdb[64+i]));
		if (rc) {
			fprintf(stderr, "Can't open database %s: %s\n", buffer2, sqlite3_errmsg(ws->localdb[64+i]));
			sqlite3_close(ws->localdb[64+i]);
			return rc;
		};
		run_statement(ws, ws->localdb[64+i], "PRAGMA cache_size=10000;");
		run_statement(ws, ws->localdb[64+i], "PRAGMA synchronous=OFF;");
	};
	// o-part
	for (i = 1; i <= 64; i++) {
		sprintf(buffer2, "%so%d.db", buffer, i);
		rc = sqlite3_open(buffer2, &(ws->localdb[128+i]));
		if (rc) {
			fprintf(stderr, "Can't open database %s: %s\n", buffer2, sqlite3_errmsg(ws->localdb[128+i]));
			sqlite3_close(ws->localdb[128+i]);
			return rc;
		};
		run_statement(ws, ws->localdb[128+i], "PRAGMA cache_size=10000;");
		run_statement(ws, ws->localdb[128+i], "PRAGMA synchronous=OFF;");
	};
	
	globaldb = config_lookup(ws->cfg, "globaldb");
	if (globaldb == NULL) {
		fprintf(stderr, "Configuration file does not contain globaldb directive.\n");
		return 1;
	};
	rc = strlen(config_setting_get_string(globaldb));
	if (rc > 500) {
		fprintf(stderr, "Configuration files 'globaldb' directive cannot exceed 500 characters\n");
		return 1;
	};
	zero_string(buffer, 512);
	strncpy(buffer, config_setting_get_string(globaldb), rc);
	rc = sqlite3_open(buffer, &(ws->matchdb));
	if (rc) {
		fprintf(stderr, "Can't open database %s: %s\n", buffer, sqlite3_errmsg(ws->matchdb));
		sqlite3_close(ws->matchdb);
		return rc;
	};
	run_statement(ws, ws->matchdb, "PRAGMA cache_size=10000;");
	run_statement(ws, ws->matchdb, "PRAGMA synchronous=OFF;");

	if (! ws->arg->initdatabases)
		run_upgrade_fixes(ws);

	return rc;
};

void add_to_database(struct workstate * ws, struct cpe_data cpe) {
	int rc = 0;
	char stmt[1024];

	// Full match

	sprintf(stmt, "select count(*) from tb_cpe_%c_%zu where cpepart = \"%c\" and cpevendor = \"%s\" and cpeproduct = \"%s\" and cpeversion = \"%s\" and cpeupdate = \"%s\" and cpeedition = \"%s\" and cpelanguage = \"%s\";", cpe.part, strlen(cpe.vendor), cpe.part, cpe.vendor, cpe.product, cpe.version, cpe.update, cpe.edition, cpe.language);

	rc = get_int_value(get_local_db(ws, cpe.part, strlen(cpe.vendor)), stmt, ws);

	if (rc == 0) {
		sprintf(stmt, "insert into tb_cpe_%c_%zu (cpepart, cpevendor, cpeproduct, cpeversion, cpeupdate, cpeedition, cpelanguage) values (\"%c\", \"%s\", \"%s\", \"%s\", \"%s\", \"%s\", \"%s\");", cpe.part, strlen(cpe.vendor), cpe.part, cpe.vendor, cpe.product, cpe.version, cpe.update, cpe.edition, cpe.language);
		run_statement(ws, get_local_db(ws, cpe.part, strlen(cpe.vendor)), stmt);

		sprintf(stmt, "select cpeid from tb_cpe_%c_%zu where cpepart = \"%c\" and cpevendor = \"%s\" and cpeproduct = \"%s\" and cpeversion = \"%s\" and cpeupdate = \"%s\" and cpeedition = \"%s\" and cpelanguage = \"%s\";", cpe.part, strlen(cpe.vendor), cpe.part, cpe.vendor, cpe.product, cpe.version, cpe.update, cpe.edition, cpe.language);
	
		rc = get_int_value(get_local_db(ws, cpe.part, strlen(cpe.vendor)), stmt, ws);
	} else {
		sprintf(stmt, "select cpeid from tb_cpe_%c_%zu where cpepart = \"%c\" and cpevendor = \"%s\" and cpeproduct = \"%s\" and cpeversion = \"%s\" and cpeupdate = \"%s\" and cpeedition = \"%s\" and cpelanguage = \"%s\";", cpe.part, strlen(cpe.vendor), cpe.part, cpe.vendor, cpe.product, cpe.version, cpe.update, cpe.edition, cpe.language);
		rc = get_int_value(get_local_db(ws, cpe.part, strlen(cpe.vendor)), stmt, ws);
	};

	sprintf(stmt, "delete from tb_binmatch where basedir = \"%s\" and filename = \"%s\";", ws->currentdir, ws->currentfile);
	run_statement(ws, ws->localdb[0], stmt);

	sprintf(stmt, "insert into tb_binmatch values ('%s', '%s', '%c', %zu, %d, 1);", ws->currentdir, ws->currentfile, cpe.part, strlen(cpe.vendor), rc);
	run_statement(ws, ws->localdb[0], stmt);
};


int file_already_processed(struct workstate * ws) {
 	int rc = 0;
	char stmt[1024];

	sprintf(stmt, "select count(*) from tb_binmatch where basedir = \"%s\" and filename = \"%s\";", ws->currentdir, ws->currentfile);

	rc = get_int_value(ws->localdb[0], stmt, ws);

	if (ws->rc != 0) {
		return 1;
	};

	return 0;
};

int display_cve_matches(void * cbobj, int argc, char **argv, char **azColName) {
	int i;
	int j;
	char * basedir = NULL;
	char * filename = NULL;
	char * cpe = NULL;
	char cpepart = 'i'; // Illegal initial value, just in case
	char * cpevendorlength = NULL;
	int year = 0;
	int sequence = 0;
	int versionOnly = 0;

	j = 0;

	for (i = 0; i < argc; i++) {
		if (strcmp(azColName[i], "basedir") == 0) {
			basedir = argv[i];
			j++;
			continue;
		};
		if (strcmp(azColName[i], "filename") == 0) {
			filename = argv[i];
			j++;
			continue;
		}
		if (strcmp(azColName[i], "year") == 0) {
			year = atoi(argv[i]);
			j++;
			continue;
		}
		if (strcmp(azColName[i], "sequence") == 0) {
			sequence = atoi(argv[i]);
			j++;
			continue;
		}
		if (strcmp(azColName[i], "cpe") == 0) {
			cpe = argv[i];
			j++;
			continue;
		}
		if (strcmp(azColName[i], "cpepart") == 0) {
			cpepart = argv[i][0];
			j++;
			continue;
		}
		if (strcmp(azColName[i], "cpevendorlength") == 0) {
			cpevendorlength = argv[i];
			j++;
			continue;
		}
		if (strcmp(azColName[i], "fullmatch") == 0) {
			versionOnly = atoi(argv[i]) == 0 ? 1 : 0;
			j++;
			continue;
		};
		fprintf(stderr, "Error: got column \"%s\" with no match.\n", azColName[i]);
	};

	if (j == 8) {
		char stmt[1024];
		int rc = 0;
		struct cpe_data cpedata;
		int length = atoi(cpevendorlength);
		struct workstate * ws = (struct workstate *) cbobj;

		sprintf(stmt, "select cpepart, cpevendor, cpeproduct, cpeversion, cpeupdate, cpeedition, cpelanguage from tb_cpe_%c_%d where cpeid = %s;", cpepart, length, cpe);
		rc = get_cpe_data(get_local_db(ws, cpepart, length), stmt, ws);
		cpedata = ws->cpebuffer;

		sprintf(stmt, "%s/%s", basedir, filename);

		show_potential_vulnerabilities(ws,  year,  sequence, stmt, cpedata, versionOnly);
	} else {
		fprintf(stderr, "Error, got %d entries back from query.\n", j);
	};

	return 0;
};

int display_cve_matches_for_cpeid(void * cbobj, int argc, char **argv, char **azColName) {
	int i;
	int cpeid = -1;
	int parentcpeid = -1;
	int rc;
	char cpepart;
	int cpelength;
	char stmt[1024];
	char * errmsg;
	struct workstate * ws = (struct workstate *) cbobj;

	for (i = 0; i < argc; i++) {
		if (strcmp(azColName[i], "cpeid") == 0) {
			cpeid = atoi(argv[i]);
			continue;
		};
		if (strcmp(azColName[i], "parentcpeid") == 0) {
			parentcpeid = atoi(argv[i]);
			continue;
		};
	};

	if ((cpeid == -1) || (parentcpeid == -1)) {
		fprintf(stderr, "Failed to query for subset of CPE identifiers, but then again, why did I reach this part of the code then?\n");
		return 1;
	};

	if (ws->rc < 65) {
		// cpeid for a-part, length = rc
		cpepart = 'a';
		cpelength = ws->rc;
	} else if (ws->rc < 129) {
		// cpeid for h-part, length = rc-64
		cpepart = 'h';
		cpelength = ws->rc - 64;
	} else if (ws->rc < 193) {
		// cpeid for o-part, length = rc-128
		cpepart = 'o';
		cpelength = ws->rc - 128;
	} else {
		fprintf(stderr, "Now this is awkward... should never reach this code ?!?\n");
		return 1;
	}

	sprintf(stmt, "select b.basedir, b.filename, c.year, c.sequence, 0 as fullmatch, b.cpe, b.cpepart, b.cpevendorlength from tb_binmatch b, tb_cve c where b.cpe = %d and c.cpe = %d and b.cpepart = c.cpepart and b.cpevendorlength = c.cpevendorlength;", parentcpeid, cpeid);

	rc = sqlite3_exec(ws->localdb[0], stmt, display_cve_matches, ws, &errmsg);
	if (rc) {
		fprintf(stderr, "Failed to add information to database: %s\n", errmsg);
		sqlite3_free(errmsg);
	};

	return 0;
}


/**
 * Process the results of the file-matching query.
 *
 * The file matching query should provide a list of possible matches. These
 * matches in the master database contain information on how to extract the
 * version of the (matching) file. This function will then use this information
 * to obtain the version from the file and store it in the database.
 */
int get_version_and_store(void * cbobj, int argc, char **argv, char **azColName) {
	int i, rc;
	char * file_name_expression = NULL;
	char * file_content_expression = NULL;
	regex_t preg;
	regmatch_t pmatch[16];
	struct cpe_data cpe_data;
	struct workstate * ws = (struct workstate *) cbobj;
	int filetype = 0;

	ws->rc=0;	// Re-init state

	for(i=0;i<argc;i++) {
		// One or more hits
		if (strcmp(azColName[i], "filematch") == 0) {
			file_name_expression = argv[i];
			continue;
		};
		if (strcmp(azColName[i], "contentmatch") == 0) {
			file_content_expression = argv[i];
			continue;
		};
		if (strcmp(azColName[i], "cpepart") == 0) {
			cpe_data.part = argv[i][0];
			continue;
		};
		if (strcmp(azColName[i], "cpevendor") == 0) {
			strncpy(cpe_data.vendor, argv[i], 64);
			continue;
		};
		if (strcmp(azColName[i], "cpeproduct") == 0) {
			strncpy(cpe_data.product, argv[i], 64);
			continue;
		};
		if (strcmp(azColName[i], "cpeversion") == 0) {
			strncpy(cpe_data.version, argv[i], 64);
			continue;
		};
		if (strcmp(azColName[i], "cpeupdate") == 0) {
			strncpy(cpe_data.update, argv[i], 64);
			continue;
		};
		if (strcmp(azColName[i], "cpeedition") == 0) {
			strncpy(cpe_data.edition, argv[i], 64);
			continue;
		};
		if (strcmp(azColName[i], "cpelanguage") == 0) {
			strncpy(cpe_data.language, argv[i], 64);
			continue;
		};
		if (strcmp(azColName[i], "filetype") == 0) {
			filetype = atoi(argv[i]);
			continue;
		};
	}

	// Verify if the filename matches file_name_expression
	rc = regcomp(&preg, file_name_expression, REG_EXTENDED);
	if (rc) {
		fprintf(stderr, "Failed to compile regular expression \"%s\"\n", file_name_expression);
		ws->rc = rc;
		return 0;
	};

	rc = regexec(&preg, ws->currentfile, 16, pmatch, 0);
	if (rc) {
		regfree(&preg);
		ws->rc = rc;
		return 0;
	};
	// Filename matches.

	// Free our memory allocations, and see if the content matches as well.
	regfree(&preg);
	rc = regcomp(&preg, file_content_expression, REG_EXTENDED);
	if (rc) {
		fprintf(stderr, "Failed to compile regular expression \"%s\"\n", file_content_expression);
		ws->rc = rc;
		return 0;
	};
	/**
	 * Here is where the various version extraction methods are supported.
	 * We currently still only support a single method (1, which is the
	 * "strings -n 3 <file>" command execution) but now we can see if we can
	 * support additional methods as well.
	 */
	if (filetype == 1) {
		char buffer[512];
		int ret;

		zero_string(buffer, 512);
		ret = strings_extract_version(ws, &preg, pmatch, &cpe_data);

		if (ret == 0) {
			add_to_database(ws, cpe_data);
			cpe_to_string(buffer, 512, cpe_data);
			fprintf(stdout, " - Found match for %s/%s:\t%s\n", ws->currentdir, ws->currentfile, buffer);
		};
	} else {
		fprintf(stderr, " ! %s/%s: The sqlite3 implementation currently doesn't support file type %d\n", ws->currentdir, ws->currentfile, filetype);
	};
	ws->rc = 0;
	return 0;
}

/**
 * Process the binary
 *
 * The binary file is provided through the workstate variable (currentfile and
 * currentdir properties).
 */
int dbimpl_process_binary(struct workstate * ws) {
	char buffer[1024];
	char * errmsg;

	int rc;

	zero_string(buffer, 1024);

	/*
	 * Query: find match in master database for the file. If match found,
	 * extract the version and store it in the local database.
	 */
	strcpy(buffer, "select v.filename as filename, v.filetype as filetype, v.filematch as filematch, v.contentmatch as contentmatch, c.cpepart as cpepart, c.cpevendor as cpevendor, c.cpeproduct as cpeproduct, c.cpeversion as cpeversion, c.cpeupdate as cpeupdate, c.cpeedition as cpeedition, c.cpelanguage as cpelanguage from tb_versionmatch v, tb_cpe c where v.cpe = c.cpeid and \"");
	strcat(buffer, ws->currentfile);
	strcat(buffer,"\" between filename and filename || \"z\";");

	rc = sqlite3_exec(ws->matchdb, buffer, get_version_and_store, ws, &errmsg);
	if (rc != SQLITE_OK) {
		fprintf(stderr, "SQL error: %s\n", errmsg);
		sqlite3_free(errmsg);
		return rc;
	};

	return 0;
};

void dbimpl_verify_installed_versus_cve(struct workstate * ws) {
	int rc = 0;
	char * errmsg;
	char stmt[1024];
	int i;
	
	sprintf(stmt, "select b.basedir, b.filename, c.year, c.sequence, 1 as fullmatch, b.cpe, b.cpepart, b.cpevendorlength from tb_binmatch b, tb_cve c where b.cpe = c.cpe and b.cpepart = c.cpepart and b.cpevendorlength = c.cpevendorlength;");

	rc = sqlite3_exec(ws->localdb[0], stmt, display_cve_matches, ws, &errmsg);
	if (rc) {
		fprintf(stderr, "Failed to add information to database: %s\n", errmsg);
		sqlite3_free(errmsg);
	};

	// Next, find cpeid's that are a subset of the selected cpeid.
	// Algorithm: perform a select on the same table (foo a, foo b)
	// and look for cpeids (that are not the same) where the core
	// fields are the same, and the non-core fields are the same as well, or where
	// at least one non-core field is different and not 0.
	for (i = 1; i <= 64; i++) {
		sprintf(stmt, "select a.cpeid as cpeid, b.cpeid as parentcpeid from tb_cpe_a_%d a, tb_cpe_a_%d b where (a.cpevendor = b.cpevendor) and (a.cpeproduct = b.cpeproduct) and (a.cpeversion = b.cpeversion) and (a.cpeid <> b.cpeid) and ( (a.cpeedition = b.cpeedition) or ( (a.cpeedition <> 1 ) and (b.cpeedition == 0) ) ) and ( (a.cpeupdate = b.cpeupdate) or ( (a.cpeupdate <> 0 ) and (b.cpeupdate == 0 ) ) ) and ( (a.cpelanguage = b.cpelanguage) or ( (a.cpelanguage <> 0) and (b.cpelanguage == 0 ) ) );", i, i);
		ws->rc = i;
		rc = sqlite3_exec(get_local_db(ws, 'a', i), stmt, display_cve_matches_for_cpeid, ws, &errmsg);
		if (rc) {
			fprintf(stderr, "Failed to query database: %s\n", errmsg);
			sqlite3_free(errmsg);
		}
		sprintf(stmt, "select a.cpeid as cpeid, b.cpeid as parentcpeid from tb_cpe_h_%d a, tb_cpe_h_%d b where (a.cpevendor = b.cpevendor) and (a.cpeproduct = b.cpeproduct) and (a.cpeversion = b.cpeversion) and (a.cpeid <> b.cpeid) and ( (a.cpeedition = b.cpeedition) or ( (a.cpeedition <> 0 ) and (b.cpeedition == 0) ) ) and ( (a.cpeupdate = b.cpeupdate) or ( (a.cpeupdate <> 0 ) and (b.cpeupdate == 0 ) ) ) and ( (a.cpelanguage = b.cpelanguage) or ( (a.cpelanguage <> 0) and (b.cpelanguage == 0 ) ) );", i, i);
		ws->rc = 64+i;
		rc = sqlite3_exec(get_local_db(ws, 'h', i), stmt, display_cve_matches_for_cpeid, ws, &errmsg);
		if (rc) {
			fprintf(stderr, "Failed to query database: %s\n", errmsg);
			sqlite3_free(errmsg);
		}
		sprintf(stmt, "select a.cpeid as cpeid, b.cpeid as parentcpeid from tb_cpe_o_%d a, tb_cpe_o_%d b where (a.cpevendor = b.cpevendor) and (a.cpeproduct = b.cpeproduct) and (a.cpeversion = b.cpeversion) and (a.cpeid <> b.cpeid) and ( (a.cpeedition = b.cpeedition) or ( (a.cpeedition <> 0 ) and (b.cpeedition == 0) ) ) and ( (a.cpeupdate = b.cpeupdate) or ( (a.cpeupdate <> 0 ) and (b.cpeupdate == 0 ) ) ) and ( (a.cpelanguage = b.cpelanguage) or ( (a.cpelanguage <> 0) and (b.cpelanguage == 0 ) ) );", i, i);
		ws->rc = 128+i;
		rc = sqlite3_exec(get_local_db(ws, 'o', i), stmt, display_cve_matches_for_cpeid, ws, &errmsg);
		if (rc) {
			fprintf(stderr, "Failed to query database: %s\n", errmsg);
			sqlite3_free(errmsg);
		}
	};
};

void dbimpl_initialize_workstate(struct workstate * ws) {
	int i = 0;

	ws->matchdb = NULL;
	ws->localdb = (sqlite3 **) calloc(sizeof(sqlite3 *), 64*3+1);
	for (i = 0; i <= 64*3; i++) {
		ws->localdb[i] = NULL;
	};
};

int check_cvecpe_in_db(struct workstate * ws, int year, int sequence, struct cpe_data cpe) {
	int rc = 0;
	char stmt[1024];

	sprintf(stmt, "select cpeid from tb_cpe_%c_%zu where cpepart = \"%c\" and cpevendor = \"%s\" and cpeproduct = \"%s\" and cpeversion = \"%s\" and cpeupdate = \"%s\" and cpeedition = \"%s\" and cpelanguage = \"%s\";", cpe.part, strlen(cpe.vendor), cpe.part, cpe.vendor, cpe.product, cpe.version, cpe.update, cpe.edition, cpe.language);

	rc = get_int_value(get_local_db(ws, cpe.part, strlen(cpe.vendor)), stmt, ws);

	if (rc == 0) {
		sprintf(stmt, "insert into tb_cpe_%c_%zu (cpepart, cpevendor, cpeproduct, cpeversion, cpeupdate, cpeedition, cpelanguage) values (\"%c\", \"%s\", \"%s\", \"%s\", \"%s\", \"%s\", \"%s\");", cpe.part, strlen(cpe.vendor), cpe.part, cpe.vendor, cpe.product, cpe.version, cpe.update, cpe.edition, cpe.language),
		run_statement(ws, get_local_db(ws, cpe.part, strlen(cpe.vendor)), stmt);
		sprintf(stmt, "select cpeid from tb_cpe_%c_%zu where cpepart = \"%c\" and cpevendor = \"%s\" and cpeproduct = \"%s\" and cpeversion = \"%s\" and cpeupdate = \"%s\" and cpeedition = \"%s\" and cpelanguage = \"%s\";", cpe.part, strlen(cpe.vendor), cpe.part, cpe.vendor, cpe.product, cpe.version, cpe.update, cpe.edition, cpe.language);
		rc = get_int_value(get_local_db(ws, cpe.part, strlen(cpe.vendor)), stmt, ws);
	};

	sprintf(stmt, "select count(*) from tb_cve where year = %d and sequence = %d and cpepart = '%c' and cpevendorlength = %zu and cpe = %d;", year, sequence, cpe.part, strlen(cpe.vendor), rc);
	rc = get_int_value(ws->localdb[0], stmt, ws);
	return rc;
};

/**
 * Store the passed CVE entry in the database
 */
int dbimpl_store_cve_in_db(struct workstate * ws, char * cveId, char * cpeId) {
	int rc = 0;
	char stmt[1024];
	int year, sequence;
	struct cpe_data cpe;

	rc = cve_to_vars(&year, &sequence, cveId);
	if (rc == 1) {
		return 0;
	};
	string_to_cpe(&cpe, cpeId);

	ws->rc = 0;
	rc = check_cvecpe_in_db(ws, year, sequence, cpe);
	if ((rc != 0) && (ws->rc != 0)) {
		return 0;
	};

	sprintf(stmt, "select cpeid from tb_cpe_%c_%zu where cpepart = \"%c\" and cpevendor = \"%s\" and cpeproduct = \"%s\" and cpeversion = \"%s\" and cpeupdate = \"%s\" and cpeedition = \"%s\" and cpelanguage = \"%s\";",  cpe.part, strlen(cpe.vendor), cpe.part, cpe.vendor, cpe.product, cpe.version, cpe.update, cpe.edition, cpe.language);
	rc = get_int_value(get_local_db(ws, cpe.part, strlen(cpe.vendor)), stmt, ws);

	sprintf(stmt, "insert into tb_cve values (%d, %d, '%c', %zu, %d);", year, sequence, cpe.part, strlen(cpe.vendor), rc);
	run_statement(ws, ws->localdb[0], stmt);

	return 1;
};


/**
 * Add versiongather entry into the database
 */
int dbimpl_add_versiongather(struct workstate * ws, struct versiongather_data vg, struct cpe_data cpe) {
	char stmt[1024];
	char stmt2[1024];
	int cpid = 0;

	sprintf(stmt, "select cpeid from tb_cpe where cpepart = \"%c\" and cpevendor = \"%s\" and cpeproduct = \"%s\" and cpeversion = \"%s\" and cpeupdate = \"%s\" and cpeedition = \"%s\" and cpelanguage = \"%s\";", cpe.part, cpe.vendor, cpe.product, cpe.version, cpe.update, cpe.edition, cpe.language);

	cpid = get_int_value(ws->matchdb, stmt, ws);

	if (cpid == 0) {
		zero_string(stmt2, 1024);
		sprintf(stmt2, "insert into tb_cpe (cpepart, cpevendor, cpeproduct, cpeversion, cpeupdate, cpeedition, cpelanguage) values (\"%c\", \"%s\", \"%s\", \"%s\", \"%s\", \"%s\", \"%s\");", cpe.part, cpe.vendor, cpe.product, cpe.version, cpe.update, cpe.edition, cpe.language);
		run_statement(ws, ws->matchdb, stmt2);

		cpid = get_int_value(ws->matchdb, stmt, ws);
	};

	if (cpid == 0) {
		fprintf(stderr, "Failed to store new data!\n");
		return 0;
	};

	zero_string(stmt2, 1024);
	sprintf(stmt2, "insert into tb_versionmatch values (\"%s\", %d, \"%s\", \"%s\", %d);", vg.filepart, vg.gathertype, vg.filematch, vg.versionexpression, cpid);
	run_statement(ws, ws->matchdb, stmt2);

	return 1;
};

/**
 * Initialize the databases
 */
int dbimpl_initialize_databases(struct workstate * ws) {
	char stmt[1024];
	int size = 1;

	// Setup of global
	sprintf(stmt, "PRAGMA foreign_keys=OFF; BEGIN TRANSACTION; DROP TABLE IF EXISTS tb_versionmatch; DROP TABLE IF EXISTS tb_cpe; CREATE TABLE tb_versionmatch (filename varchar(64), filetype smallint, filematch varchar(128), contentmatch varchar(256), cpe int); CREATE INDEX vmidx ON tb_versionmatch (filename); CREATE TABLE tb_cpe (cpeid integer primary key, cpepart char(1), cpevendor varchar(64), cpeproduct varchar(64), cpeversion varchar(64), cpeupdate varchar(64), cpeedition varchar(64), cpelanguage varchar(64)); COMMIT;");
	run_statement(ws, ws->matchdb, stmt);

	// Setup of locals
	zero_string(stmt, 1024);
	sprintf(stmt, "PRAGMA foreign_keys=OFF; BEGIN TRANSACTION; DROP TABLE IF EXISTS tb_binmatch; DROP TABLE IF EXISTS tb_cve; CREATE TABLE tb_binmatch ( basedir varchar(256), filename varchar(256), cpepart char(1), cpevendorlength int, cpe int, fullmatch int); CREATE TABLE tb_cve ( year smallint, sequence int, cpepart char(1), cpevendorlength int, cpe int); CREATE INDEX cveidx ON tb_cve (year, sequence); CREATE INDEX binmatchidx on tb_binmatch (cpe, cpepart, cpevendorlength); COMMIT;");
	run_statement(ws, ws->localdb[0], stmt);

	for (size = 1; size <= 64; size++) {
		zero_string(stmt, 1024);
		sprintf(stmt, "DROP TABLE IF EXISTS tb_cpe_a_%d; CREATE TABLE tb_cpe_a_%d (cpeid integer primary key, cpepart char(1), cpevendor char(64), cpeproduct char(64), cpeversion char(64), cpeupdate char(64), cpeedition char(64), cpelanguage char(64)); CREATE INDEX cpe_a_%d_idx on tb_cpe_a_%d (cpevendor, cpeproduct, cpeversion, cpeid, cpeedition, cpeupdate, cpelanguage);", size, size, size, size);
		run_statement(ws, get_local_db(ws, 'a', size), stmt);
	
		zero_string(stmt, 1024);
		sprintf(stmt, "DROP TABLE IF EXISTS tb_cpe_h_%d; CREATE TABLE tb_cpe_h_%d (cpeid integer primary key, cpepart char(1), cpevendor char(64), cpeproduct char(64), cpeversion char(64), cpeupdate char(64), cpeedition char(64), cpelanguage char(64)); CREATE INDEX cpe_h_%d_idx on tb_cpe_h_%d (cpevendor, cpeproduct, cpeversion, cpeid, cpeedition, cpeupdate, cpelanguage);", size, size, size, size);
		run_statement(ws, get_local_db(ws, 'h', size), stmt);
		
		zero_string(stmt, 1024);
		sprintf(stmt, "DROP TABLE IF EXISTS tb_cpe_o_%d; CREATE TABLE tb_cpe_o_%d (cpeid integer primary key, cpepart char(1), cpevendor char(64), cpeproduct char(64), cpeversion char(64), cpeupdate char(64), cpeedition char(64), cpelanguage char(64)); CREATE INDEX cpe_o_%d_idx on tb_cpe_o_%d (cpevendor, cpeproduct, cpeversion, cpeid, cpeedition, cpeupdate, cpelanguage);", size, size, size, size);
		run_statement(ws, get_local_db(ws, 'o', size), stmt);
	};

	return 1;
};

/**
 * Report on detected software/versions
 */
void dbimpl_report_installed(struct workstate * ws, int showfiles) {
	char stmt[1024];
	char * errmsg;
	int size;
	int rc;

	for (size = 1; size <= 64; size++) {
		sprintf(stmt, "SELECT DISTINCT cpepart, cpevendorlength, cpe FROM tb_binmatch WHERE cpepart = 'a' AND cpevendorlength = %d;", size);
		ws->rc = showfiles;
		rc = sqlite3_exec(ws->localdb[0], stmt, get_cpelist, ws, &errmsg);
		if (rc) {
			fprintf(stderr, "Failed to get list of installed software: %s\n", errmsg);
			sqlite3_free(errmsg);
		};

		sprintf(stmt, "SELECT DISTINCT cpepart, cpevendorlength, cpe FROM tb_binmatch WHERE cpepart = 'h' AND cpevendorlength = %d;", size);
		ws->rc = showfiles;
		rc = sqlite3_exec(ws->localdb[0], stmt, get_cpelist, ws, &errmsg);
		if (rc) {
			fprintf(stderr, "Failed to get list of installed software: %s\n", errmsg);
			sqlite3_free(errmsg);
		};

		sprintf(stmt, "SELECT DISTINCT cpepart, cpevendorlength, cpe FROM tb_binmatch WHERE cpepart = 'o' AND cpevendorlength = %d;", size);
		ws->rc = showfiles;
		rc = sqlite3_exec(ws->localdb[0], stmt, get_cpelist, ws, &errmsg);
		if (rc) {
			fprintf(stderr, "Failed to get list of installed software: %s\n", errmsg);
			sqlite3_free(errmsg);
		};


	};

};

/**
 * Initialization for storing CVE data in the database.
 *
 * The initialization function is meant to allow implementations
 * to optimize their functions (like starting a transaction scope).
 */
void dbimpl_store_cve_in_db_init(struct workstate * ws) {
	// do nothing
};

/**
 * Exit method for finalizing the storage of CVE data in the database.
 *
 * The exit method is meant to allow implementations to optimize their functions
 * (like committing at the transaction scope).
 */
void dbimpl_store_cve_in_db_exit(struct workstate * ws) {
	// do nothing
};

/**
 * Initialize arguments structure with database-specific fields
 */
void dbimpl_initialize_arguments(struct arguments * arg) {
	// do nothing
};
